use crate::admin::domain::{AdminApp, AdminModel};
use axum::Router;
use serde_json::Value;
use std::collections::HashMap;
use std::sync::Arc;

pub type AdminState = Arc<AdminRegistry>;

pub trait AdminRepository: Send + Sync {
    fn get_item(&self, id: usize) -> Option<Value>;
    fn get_list(&self) -> Value;
}

// main registry.
pub struct AdminRegistry {
    base_path: String,
    apps: HashMap<String, internal::AdminApp>,
    models: HashMap<String, internal::AdminModel>,
    repositories: HashMap<String, Box<dyn AdminRepository>>,
}

impl AdminRegistry {
    pub fn new(base_path: &str) -> Self {
        AdminRegistry {
            base_path: base_path.to_owned(),
            apps: HashMap::new(),
            models: HashMap::new(),
            repositories: HashMap::new(),
        }
    }

    pub fn generate_router<T>(&self) -> Router<T>
    where
        T: Clone,
        T: Send,
        T: Sync,
        T: 'static,
        crate::service::templates::Templates: axum::extract::FromRef<T>,
        AdminState: axum::extract::FromRef<T>,
    {
        routing::generate_routes::<T>(&self)
    }

    pub fn get_apps(&self) -> Vec<AdminApp> {
        self.apps
            .iter()
            .map(|(key, node)| self.get_app(key, node))
            .collect()
    }

    fn get_app(&self, name: &str, node: &internal::AdminApp) -> AdminApp {
        let label = node.label.as_ref().unwrap_or(&String::new()).clone();
        AdminApp {
            name: name.to_owned(),
            app_label: label,
            app_url: name.to_owned().to_lowercase(),
            models: vec![],
        }
    }

    pub fn register_app(&mut self, name: &str, app_label: &str, app_url: &str) {
        self.apps.insert(
            name.to_owned(),
            internal::AdminApp {
                object_name: name.to_owned(),
                label: Some(app_label.to_owned()),
            },
        );
    }

    pub fn get_models(&self, app_name: &str) -> Vec<AdminModel> {
        vec![]
    }

    pub fn get_model(&self, app_name: &str, model_name: &str) -> Option<AdminModel> {
        let full_model_name = format!("{}.{}", app_name, model_name);
        let internal_model = self.models.get(&full_model_name)?;

        // unfinished: we need to think about model_name vs. model_id vs. entry_id, as "name" is ambiguous.
        Some(AdminModel {
            object_name: internal_model.model_name.clone(),
            name: internal_model.model_name.clone(),
            admin_url: "".to_owned(),
            view_only: false,
            add_url: None,
        })
    }

    fn register_model_config(&mut self, model: config::AdminModelConfig) -> Result<String, String> {
        let local_config = internal::AdminModel::from(model);
        if local_config.model_name.is_empty() {
            return Err("No model name".to_owned());
        }
        let local_config_name = format!("{}.{}", local_config.app_name, local_config.model_name);
        if self.models.contains_key(&local_config_name) {
            return Err(format!("Model {} already exists", local_config_name));
        }
        if let Some(local_config) = self.models.insert(local_config_name, local_config) {
            return Ok(local_config.model_name);
        }
        Err("Model could not be added!".to_owned())
    }

    pub fn register_model<R: AdminRepository + 'static>(
        &mut self,
        model: config::AdminModelConfig,
        repository: R,
    ) -> Result<(), String> {
        let model_name = self.register_model_config(model)?;
        self.repositories.insert(model_name, Box::new(repository));
        Ok(())
    }
}

pub mod config {
    use axum::routing::MethodRouter;

    // user uses this configuration object to register another model.
    pub struct AdminModelConfig {
        pub name: String,
        pub app_name: String,

        pub custom_index_handler: Option<MethodRouter>,
    }
}

mod internal {
    #[derive(Clone)]
    pub struct AdminApp {
        pub object_name: String,
        pub label: Option<String>,
    }

    #[derive(Clone)]
    pub struct AdminModel {
        pub app_name: String,
        pub model_name: String,
    }

    impl From<super::config::AdminModelConfig> for AdminModel {
        fn from(value: super::config::AdminModelConfig) -> Self {
            AdminModel {
                app_name: value.app_name,
                model_name: value.name,
            }
        }
    }
}

mod routing {
    use axum::{routing::get, Router};

    pub fn generate_routes<T>(registry: &super::AdminRegistry) -> Router<T>
    where
        T: Clone,
        T: Send,
        T: Sync,
        T: 'static,
        crate::service::templates::Templates: axum::extract::FromRef<T>,
        super::AdminState: axum::extract::FromRef<T>,
    {
        let mut router: Router<T> = Router::new();
        let apps = registry.get_apps();
        router = router.route(
            &format!("/{}", registry.base_path),
            get(crate::admin::views::index),
        );
        for app in apps {
            let app_route = format!("/{}/{}", registry.base_path, app.app_url);

            // Add a route for the app
            router = router.route(
                &app_route,
                get(move || async move { format!("Admin panel for {}", app.app_label) }),
            );

            // Add routes for each model in the app
            /*for model in data_node.models.iter() {
                let model_route = format!("{}/{}", app_route, model.name);
                router = router.route(
                    &model_route,
                    get(move || async move { format!("Model page for {}", model.name) }),
                );
            }*/
        }

        router
    }

    fn define_example_data(registry: &mut super::AdminRegistry) {
        /*let models = vec![
            AdminModel {
                name: "User".to_owned(),
                object_name: "User".to_owned(),
                admin_url: "/admin/users/user".to_owned(),
                view_only: false,
                add_url: Some("/admin/users/user/add".to_owned()),
            },
            AdminModel {
                name: "Group".to_owned(),
                object_name: "Group".to_owned(),
                admin_url: "/admin/users/group".to_owned(),
                view_only: false,
                add_url: None,
            },
            AdminModel {
                name: "Permission".to_owned(),
                object_name: "Permission".to_owned(),
                admin_url: "/admin/users/permission".to_owned(),
                view_only: true,
                add_url: None,
            },
        ];*/
        registry.register_app("auth", "Authorities", "auth");
    }
}
